#!/usr/bin/env python3
# -*- coding:utf-8 -*-

import argparse
import os.path
from sprp.core.alg import *
from sprp.export.geojson import *
from sprp.export.shapefile import *
from sprp.export.ply import *
from sprp.export.las import *
from sprp.export.txt import *

def main():
    parser = argparse.ArgumentParser(description="A simple photogrammetry route planner for UAV.")

    parser.add_argument( '--filetype', '-t', help="Format of export file")

    parser.add_argument('-l',  '--length', help="Length of camera",type=str,required=True)
    parser.add_argument('-w',  '--width', help="Width of camera",type=int,required=True)
    parser.add_argument('-p',  '--pixelsize', help="Pixel size of camera",type=float,required=True)
    parser.add_argument('-f',  '--focuslength', help="Focus length of camera",type=float,required=True)

    parser.add_argument('-c',  '--courseoverlap', help="Overlap of course orientation",type=float,required=True)
    parser.add_argument('-s',  '--sidewiseoverlap', help="Overlap of sidewise orientation",type=float,required=True)
    parser.add_argument('-g',  '--gsd', help="GSD",type=float,required=True)

    parser.add_argument('--path', help="File path to save",required=True)
    parser.add_argument('--name', help="File name",required=True)

    parser.add_argument('-d','--wkt', help="Geometry",type=str,required=True)

    args = parser.parse_args()

    sc = SimplePolygonCalculator(args.wkt,
                                **{
                                "cameraWidth": int(float(args.length)),
                                "cameraHeight":int(float(args.width)),
                                "focusLength":float(args.focuslength),
                                "pixelSize":float(args.pixelsize),
                                "gsd":float(args.gsd),
                                "flightSpeed":80,
                                "courseOverlap":float(args.courseoverlap),
                                "sidewiseOverlap":float(args.sidewiseoverlap), }
            )
    result = sc.calculate()

    # 没有指定文件格式的情况下，默认为geojson
    #print('FileType:', args.filetype)
    if args.filetype is None or args.filetype.lower() == 'geojson':
        exportor = GeoJsonExportor(os.path.join(args.path, args.name)+'.json')
        exportor.save(sc)
    elif args.filetype.lower() == 'shapefile':
            exportor = ShapefileExporter(args.path, args.name)
            exportor.save(sc)
    elif args.filetype.lower() == 'ply':
            exportor = PlyExportor(os.path.join(args.path, args.name)+'.ply')
            exportor.save(sc)
    elif args.filetype.lower() == 'las':
            exportor = LasExportor(os.path.join(args.path, args.name)+'.las')
            exportor.save(sc)
    elif args.filetype.lower() == 'txt':
            exportor = TxtExportor(os.path.join(args.path, args.name)+'.txt')
            exportor.save(sc)

if __name__ == "__main__":
    main()


#!/usr/bin/env python3
# -*- coding:utf-8 -*-

from flask_cors import CORS
from flask import Flask, render_template,request,jsonify, send_file
import os,json,tempfile,uuid
import zipfile

from sprp.core.alg import *
from sprp.export.geojson import *
from sprp.export.shapefile import *

app = Flask(__name__)
#CORS(app)

@app.route('/')
def index():
    return render_template('index.html')

@app.errorhandler(404)
def error_404(error):
    return "对不起，这个地址未开放！"

@app.route('/api/sprp/simple')
def simple():
    startx=request.args.get("startx")
    starty=request.args.get("starty")
    endx=request.args.get("endx")
    endy=request.args.get("endy")

    ssc = SimpleStripCalculator(float(startx), float(starty),
                                float(endx), float(endy),
                            0,0, 
                            **{
                            "cameraWidth": 4000,
                            "cameraHeight":3000,
                            "focusLength":35,
                            "pixelSize":2,
                            "gsd":0.05,
                            "flightSpeed":80,
                            "courseOverlap":0.8,
                            "sidewiseOverlap":0.6, }
         )

    result = ssc.calculate()

    gje = GeoJsonExportor()
    gje.save(ssc)

    return "{}".format(gje.geojson)


@app.route('/api/sprp/polygon')
def polygon():
    polygon_wkt = request.args.get("wkt")
    focusLength = request.args.get("focusLength")
    pixelSize = request.args.get("pixelSize")
    cameraWidth = request.args.get("cameraWidth")
    cameraHeight = request.args.get("cameraHeight")
    courseOverlap = request.args.get("courseOverlap")
    sidewiseOverlap = request.args.get("sidewiseOverlap")
    gsd = request.args.get("gsd")
    filetype = request.args.get("filetype")
    user_filename = request.args.get("filename")
    
    user_filename = user_filename if (user_filename != None 
                    or user_filename == '') else "lxy"

    # print(">>> ", filetype)
   
    #try:
    # 生成geojson串
    if filetype is None:
        spc = SimplePolygonCalculator(polygon_wkt,
                                **{
                                "cameraWidth": int(float(cameraWidth)),
                                "cameraHeight":int(float(cameraHeight)),
                                "focusLength":float(focusLength),
                                "pixelSize":float(pixelSize),
                                "gsd":float(gsd),
                                "flightSpeed":80,
                                "courseOverlap":float(courseOverlap),
                                "sidewiseOverlap":float(sidewiseOverlap), }
            )

        result = spc.calculate()

        gje = GeoJsonExportor()
        gje.save(spc)

        res = {
            "status":'success',
            "data":json.loads(gje.geojson)
        } 
        return jsonify(res)
    else: # 发送文件
        spc = SimplePolygonCalculator(polygon_wkt,
                                **{
                                "cameraWidth": int(float(cameraWidth)),
                                "cameraHeight":int(float(cameraHeight)),
                                "focusLength":float(focusLength),
                                "pixelSize":float(pixelSize),
                                "gsd":float(gsd),
                                "flightSpeed":80,
                                "courseOverlap":float(courseOverlap),
                                "sidewiseOverlap":float(sidewiseOverlap), }
            )

        result = spc.calculate()
        
        if filetype == 'GeoJSON':
            file_ext = ".json"
            filename = str(uuid.uuid4()) 
            filepath = os.path.join(tempfile.gettempdir(), filename) + file_ext
            # print("GEOJSON", filepath)
            gje = GeoJsonExportor(filepath)
            gje.save(spc)
            mime='application/json,application/octet-stream'
        else:
            filename = "{}".format(uuid.uuid4()) 
            filepath = tempfile.gettempdir()
            #print("ZIP文件：{}\n,{}\n{}\n---\n".format(filetype, filepath,filename))
            sfe = ShapefileExporter(filepath, filename)
            sfe.save(spc)
            
            filepath = os.path.join(filepath,filename)
            
            # 压缩文件
            _files = []
            list_files = os.listdir(filepath)
            
            for f in list_files:
                #print("LIST FILE",f)
                if os.path.isfile(os.path.join(filepath,f)):
                    _files.append(f)
            
            zip_file = zipfile.ZipFile(filepath + ".zip", 
                                        'w', zipfile.ZIP_DEFLATED)
            for f in _files:
                #print("FILE:", f)
                zip_file.write(os.path.join(filepath,f), f)
            zip_file.close()
            filepath = filepath + '.zip'
            mime = 'application/zip,application/octet-stream'

        res = {
            "status":'sucess',
            "data":filepath
            }
        #return jsonify(res)
        return send_file(filepath, mimetype=mime,
                    as_attachment=True, attachment_filename=user_filename)
    
    # except Exception as e:
    #     print(e)
    #     res = {
    #         "status":'failed',
    #         "data":str(e)
    #     } 
    #     return jsonify(res)
        

def main():
    port = os.environ.get("SPRP_PORT", "8000")
    app.run(host='0.0.0.0', port=int(port), debug=True)

if __name__ == "__main__":
    main()


"""
A simple photogrammetry routeing planner for UAV ,

Requirements: Python 3.6+.

Contact:  Xiangyong Luo <solo_lxy@126.com>

BSD 2-Clause License

Copyright (c) 2021, luoxiangyong
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
import pyproj
import math,os,shutil
from osgeo import ogr,osr

from ..core.alg  import *

class ShapefileExporter(SimpleExportor):
    def __init__(self,path,basename):
        super().__init__()
        self.name = "ESRI Shapefile Exporter"

        self.path = path
        self.basename = basename

    def savePointsFile(self,calculator):
        ########################################################################
        # 创建点文件
        
        driver = ogr.GetDriverByName('ESRI Shapefile')
        path = os.path.join(self.path,self.basename,
                            "{}-points.shp".format(self.basename))
        #print(path)
        dataSource = driver.CreateDataSource(path)
        
        spatialReference = osr.SpatialReference()
        spatialReference.SetWellKnownGeogCS('WGS84')

        layer = dataSource.CreateLayer("layer", spatialReference)

        field = ogr.FieldDefn("ID", ogr.OFTInteger)
        field.SetWidth(4)
        layer.CreateField(field)
        field = ogr.FieldDefn("NAME", ogr.OFTString)
        field.SetWidth(20)
        layer.CreateField(field)
        field = ogr.FieldDefn("LINE", ogr.OFTString)
        field.SetWidth(20)
        layer.CreateField(field)
        field = ogr.FieldDefn("Z", ogr.OFTReal)
        field.SetWidth(20)
        field.SetPrecision(5)
        layer.CreateField(field)
        ########################################################################

        # 写入点
        id = 0
        lineIndex = 0
        
        for line in calculator.points:
            lineIndex = lineIndex + 1
            id = 0
            for p in line:
                id = id + 1
                name = "{}".format(id)
                lineName = "{}".format(lineIndex)
                wkt = "POINT({} {})".format(p[0],p[1])
                #print(wkt)
                geometry = ogr.CreateGeometryFromWkt(wkt)
                feature = ogr.Feature(layer.GetLayerDefn())
                feature.SetGeometry(geometry)
                feature.SetField("ID", id)
                feature.SetField("NAME", name)
                feature.SetField("LINE", lineName)
                feature.SetField("Z", calculator.flight_height())
                layer.CreateFeature(feature)

            calculator.set_progress_value(lineIndex, len(calculator.points), 
                            "Save shapefile for points:{}".format(lineIndex))

    def savePolygonsFile(self, calculator):
        ########################################################################
        # 创建每个点对应的多边形文件
        
        driverPointPloygon = ogr.GetDriverByName('ESRI Shapefile')
        path = os.path.join(self.path,
                    self.basename,"{}-polygons.shp".format(self.basename))
        #print(path)
        dataSourcePointPloygon = driverPointPloygon.CreateDataSource(path)
        spatialReference = osr.SpatialReference()
        spatialReference.SetWellKnownGeogCS('WGS84')

        layerPointPloygon = dataSourcePointPloygon.CreateLayer("layer", 
                                                spatialReference)

        field = ogr.FieldDefn("ID", ogr.OFTInteger)
        field.SetWidth(4)
        layerPointPloygon.CreateField(field)

        # 写入点对应的多边形
        idPolygon = 0
        lineIndex = 0
        #print("Total Line: {}".format(len(areaPoints)))
        for line in calculator.points:
            lineIndex = lineIndex + 1
            
            for p in line:
                idPolygon =  idPolygon + 1
                name = "{}".format(id)
                lineName = "{}".format(lineIndex)
                rect  = calculator.calculate_footprint_from(p)
                wkt = "POLYGON(({} {},{} {},{} {},{} {},{} {}))".format(
                    rect[0][0],rect[0][1],
                    rect[1][0],rect[1][1],
                    rect[2][0],rect[2][1],
                    rect[3][0],rect[3][1],
                    rect[0][0],rect[0][1],
                    )
                #print("POINT({},{})".format(p[0],p[1]))
                #print(wkt)
                geometry = ogr.CreateGeometryFromWkt(wkt)
                feature = ogr.Feature(layerPointPloygon.GetLayerDefn())
                feature.SetGeometry(geometry)
                feature.SetField("ID", idPolygon)
                # feature.SetField("NAME", name)
                # feature.SetField("LINE", lineName)
                layerPointPloygon.CreateFeature(feature)
            
            calculator.set_progress_value(lineIndex, len(calculator.points), 
                            "Save shapefile for polygons:{}".format(lineIndex))

    def saveLinesFile(self, calculator):
        driver = ogr.GetDriverByName('ESRI Shapefile')
        path = os.path.join(self.path,self.basename,
                            "{}-lines.shp".format(self.basename))
        #print(path)
        dataSource = driver.CreateDataSource(path)
        
        spatialReference = osr.SpatialReference()
        spatialReference.SetWellKnownGeogCS('WGS84')

        layer = dataSource.CreateLayer("line layer", spatialReference)

        field = ogr.FieldDefn("ID", ogr.OFTInteger)
        field.SetWidth(4)
        layer.CreateField(field)
        
        field = ogr.FieldDefn("LINE", ogr.OFTString)
        field.SetWidth(20)
        layer.CreateField(field)
        ########################################################################

        # 写入点
        id = 0
        lineIndex = 0
        
        for line in calculator.lines:
            lineIndex = lineIndex + 1
            id = id + 1

            name = "{}".format(id)
            lineName = "{}".format(lineIndex)
            wkt = "LineString({} {}, {} {})".format(line[0],line[1],
                                                    line[2],line[3])
            # print(wkt)
            geometry = ogr.CreateGeometryFromWkt(wkt)
            feature = ogr.Feature(layer.GetLayerDefn())
            feature.SetGeometry(geometry)
            feature.SetField("ID", id)
            feature.SetField("LINE", lineName)
            layer.CreateFeature(feature)

            calculator.set_progress_value(lineIndex, len(calculator.points), 
                            "Save shapefile for lines:{}".format(lineIndex))

    def save(self,calculator):
        filename_points = os.path.join(self.path,self.basename)
        if os.path.exists(filename_points):
            shutil.rmtree(filename_points)
        os.mkdir(filename_points)

        self.savePointsFile(calculator)
        self.saveLinesFile(calculator)
        self.savePolygonsFile(calculator)

        return True


"""
A simple photogrammetry routeing planner for UAV

Requirements: Python 3.6+.

Contact:  Xiangyong Luo <solo_lxy@126.com>

BSD 2-Clause License

Copyright (c) 2021, luoxiangyong
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
from shapely import geometry, wkt
import pyproj
import math
import os
import shutil
from osgeo import ogr, osr


class SimpleExportor:
    def __init__(self):
        self.name = "Simple Exporter"

    def save(self, calculator):
        return False


class GeojsonExporter(SimpleExportor):
    def __init__(self):
        super().__init__()
        self.name = "Geojson Exporter"

    def save(self, calculator):
        return False


###############################################################################
class SimpleCalculator:
    """
    本类提供简单航摄区域自动曝光点设计的支持
    """

    def __init__(self, **kwargs):
        self.cameraWidth = kwargs.get('cameraWidth', 3000)
        self.cameraHeight = kwargs.get('cameraHeight', 2000)
        self.focusLength = kwargs.get('focusLength', 35)
        self.pixelSize = kwargs.get('pixelSize', 2.0)
        self.gsd = kwargs.get('gsd', 35)
        self.flightSpeed = kwargs.get('flightSpeed', 80)
        self.courseOverlap = kwargs.get('courseOverlap', 0.8)
        self.sidewiseOverlap = kwargs.get('sidewiseOverlap', 0.6)

        self.courseline = (1 - self.courseOverlap) * \
            self.cameraHeight * self.gsd
        self.sidewiseline = (1 - self.sidewiseOverlap) * \
            self.cameraWidth * self.gsd

        # 存储每条航线的起点与终点 （startx, starty, endx, endy）
        self._lines = None

        # 存储整个区域的最终设计点，列表的列表
        self._points = None

        # 当前的设计航像
        self.courseAngle = None

        self.cb = None

        self.currentProgressValue = 0
        self.totalProgressValue = 100

    @property
    def points(self):
        return self._points

    @property
    def lines(self):
        return self._lines

    def flight_height(self):
        return 1000 * self.gsd * self.focusLength / self.pixelSize

    def set_pogress_callback(self, cb):
        self.cb = cb

    def set_progress_value(self, cur, total, msg):
        self.currentProgressValue = cur
        self.totalProgressValue = total
        self.emit_progress(msg)

    def emit_progress(self, msg):
        if self.cb:
            self.cb(self.currentProgressValue, self.totalProgressValue, msg)

    def __str__(self):
        resstr = """
            cameraWidth:{},
            cameraHeight:{},
            focusLength:{},
            pixelSize:{},
            gsd:{},
            flightSpeed:{},
            courseOverlap:{},
            sidewiseOverlap:{}
            courseline:{},
            sidewiseline:{}
        """.format(self.cameraWidth, self.cameraHeight,
                   self.focusLength, self.pixelSize,
                   self.gsd, self.flightSpeed,
                   self.courseOverlap, self.sidewiseOverlap,
                   self.courseline, self.sidewiseline)

        return resstr

    def stastics(self):
        if self._points is not None:
            geod = pyproj.Geod(ellps="WGS84")

            pointCount = 0
            distance = 0

            for line in self._points:
                # print(line)
                pointCount = pointCount + len(line)
                p1 = line[0]
                p2 = line[-1]

                # print("Point:", p1,p2)

                forwardAngle, backwardAngle, distanceTmp = geod.inv(
                    p1[0], p1[1], p2[0], p2[1])
                distance = distance + distanceTmp

            return {
                "flightHeight": self.flight_height(),
                "couselineCount": len(self._lines),
                "pointCount": pointCount,
                "distance": distance / 1000,
                "workingTime": distance / self.flightSpeed
            }
        else:
            return None

    def caculate_line(self, startx, starty, endx, endy):
        geod = pyproj.Geod(ellps="WGS84")
        forwardAngle, backwardAngle, distance = geod.inv(
            startx, starty, endx, endy)
        stationCount = math.floor(distance / self.courseline)
        wishedDistance = self.courseline * (stationCount + 1)
        wished_endx, wished_endy, tempAngle = geod.fwd(
            startx, starty, forwardAngle, wishedDistance)

        points = geod.npts(startx, starty, wished_endx,
                           wished_endy, stationCount - 1)

        results = []
        results.append((startx, starty))
        results.extend(points)
        results.append((wished_endx, wished_endy))

        self.courseAngle = forwardAngle

        return results, forwardAngle

    def calculate(self):
        return False

    def calculate_footprint_from(self, point):
        """
        :brief 从点和指定的角度计算地面覆盖的矩形(footprint)

        :param point: 指定点
        :param angle: 航线方向
        :param iwidth: 图像长度
        :param iheight: 图像高度
        :param gsd: 地面分辨率

        :return: 返回地面覆盖的矩形的四脚点坐标
        """
        width = self.cameraWidth * self.gsd
        height = self.cameraHeight * self.gsd

        imgAngle = math.atan(self.cameraWidth*1.0 /
                             self.cameraHeight) * 180/math.pi

        geod = pyproj.Geod(ellps="WGS84")

        # 矩形的对角线长
        distance = math.sqrt(math.pow(width, 2) + math.pow(height, 2))

        # 计算右上角点
        angleTR = self.courseAngle - imgAngle
        longTR, latTR, tmpAngle = geod.fwd(
            point[0], point[1], angleTR, distance/2)

        # 计算右下角点
        angleBR = self.courseAngle + imgAngle
        longBR, latBR, tmpAngle = geod.fwd(
            point[0], point[1], angleBR, distance/2)

        # 计算左下角点
        angleBL = angleTR + 180
        longBL, latBL, tmpAngle = geod.fwd(
            point[0], point[1], angleBL, distance/2)

        # 计算左上角点
        angleTL = angleBR + 180
        longTL, latTL, tmpAngle = geod.fwd(
            point[0], point[1], angleTL, distance/2)

        result = []
        result.append((longTR, latTR))
        result.append((longBR, latBR))
        result.append((longBL, latBL))
        result.append((longTL, latTL))
        # 多边形闭合
        result.append((longTR, latTR))

        return result


###############################################################################
class SimpleLineCalculator(SimpleCalculator):
    """
    本类提供对线性的简单航摄区域自动曝光点设计的支持
    """

    def __init__(self, startx, starty, endx, endy, **params):
        super().__init__(**params)

        self.startx = startx
        self.starty = starty
        self.endx = endx
        self.endy = endy

    def set_line(self, startx, starty, endx, endy):
        self.startx = startx
        self.starty = starty
        self.endx = endx
        self.endy = endy

    def calculate(self):

        if self.startx and self.starty and self.endx and self.endy:
            self._points = []
            line = (self.startx, self.starty, self.endx, self.endy)
            self._lines = []

            linePointsResult, forwardAngle = self.caculate_line(*line)
            self._points.append(linePointsResult)
            self._lines.append(line)

            self.currentProgressValue = 100
            self.emit_progress("process the line:1")

            return True
        else:
            return False


class SimpleStripCalculator(SimpleCalculator):
    """
    本类提供对条带区域的简单航摄区域自动曝光点设计的支持
    """

    def __init__(self, startx, starty, endx, endy,
                 leftExpand, rightExpand, **params):
        super().__init__(**params)
        self.leftExpand = leftExpand if leftExpand else 0
        self.rightExpand = rightExpand if rightExpand else 0

        self.startx = startx
        self.starty = starty
        self.endx = endx
        self.endy = endy

    def set_line(self, startx, starty, endx, endy):
        self.startx = startx
        self.starty = starty
        self.endx = endx
        self.endy = endy

    def __str__(self):
        resstr = """
            cameraWidth:{},
            cameraHeight:{},
            focusLength:{},
            pixelSize:{},
            gsd:{},
            flightSpeed:{},
            courseOverlap:{},
            sidewiseOverlap:{},
            courseline:{},
            sidewiseline:{}
            leftExpand:{},
            rightExpand:{}
        """.format(self.cameraWidth, self.cameraHeight,
                   self.focusLength, self.pixelSize,
                   self.gsd, self.flightSpeed, self.courseOverlap,
                   self.sidewiseOverlap,
                   self.courseline, self.sidewiseline,
                   self.leftExpand, self.rightExpand)

        return resstr

    def calculate(self):
        if self.startx and self.starty and self.endx and self.endy:
            self._points = []
            self._lines = []
            lineStardEndPoints = []

            ########

            geod = pyproj.Geod(ellps="WGS84")
            angle, backAngle, distanceTmp = geod.inv(
                self.startx, self.starty, self.endx, self.endy)

            long = self.startx
            lat = self.starty
            for index in range(self.leftExpand):
                long, lat, tmpAngle = geod.fwd(
                    long, lat, angle-90, self.sidewiseline)
                e_long, e_lat, tempAngle = geod.fwd(
                    long, lat, angle, distanceTmp)
                lineStardEndPoints.append((long, lat, e_long, e_lat))

            lineStardEndPoints.append(
                (self.startx, self.starty, self.endx, self.endy))

            long = self.startx
            lat = self.starty
            for index in range(self.rightExpand):
                long, lat, tmpAngle = geod.fwd(
                    long, lat, angle+90, self.sidewiseline)
                e_long, e_lat, tempAngle = geod.fwd(
                    long, lat, angle, distanceTmp)
                lineStardEndPoints.append((long, lat, e_long, e_lat))
            #######

            self.totalProgressValue = len(lineStardEndPoints)
            for line in lineStardEndPoints:
                linePointsResult, forwardAngle = self.caculate_line(*line)
                self._points.append(linePointsResult)
                self._lines.append(line)

                self.currentProgressValue = self.currentProgressValue + 1
                self.emit_progress("process the line:{}".format(
                    self.currentProgressValue))

            return True
        else:
            return False


class SimplePolygonCalculator(SimpleCalculator):
    """
    本类提供对多边形区域的简单航摄区域自动曝光点设计的支持
    """

    def __init__(self, wkt_polygon, **params):
        super().__init__(**params)

        self.poly = wkt.loads(wkt_polygon)
        # print("Before Orient:", self.poly.wkt)
        self.poly = geometry.polygon.orient(self.poly, 1.0)
        # print("After Orient:", self.poly.wkt)

        rect = self.poly.minimum_rotated_rectangle
        rect_coords = list(rect.exterior.coords)

        # 获取最佳包围矩形的三个点
        p1 = rect_coords[0]
        p2 = rect_coords[1]
        p4 = rect_coords[3]

        # 分别计算第一个点到邻近两个点的距离
        geod = pyproj.Geod(ellps="WGS84")

        # distance1 代表与第二个点的距离，CCW
        angle1, backAngle1, distance1 = geod.inv(p1[0], p1[1], p2[0], p2[1])
        # distance1 代表与第4个点的距离
        angle2, backAngle2, distance2 = geod.inv(p1[0], p1[1], p4[0], p4[1])

        #print(angle1, backAngle1, distance1)
        #print(angle2, backAngle2, distance2)

        angle1 = angle1 if angle1 > 0 else angle1 + 360
        angle2 = angle2 if angle2 > 0 else angle2 + 360

        #print(angle1, backAngle1, distance1)
        #print(angle2, backAngle2, distance2)

        # 确定使用哪个方向上的点为我所用
        self.point_first = p1
        self.point_final = p2 if distance1 > distance2 else p4
        distance_final = distance1 if distance1 < distance2 else distance2
        self.angle_final_added = 0

        directionLeft = True
        if p1[0] < p2[0]:
            if distance2 > distance1:
                directionLeft = False
            else:
                directionLeft = True
        else:
            if distance2 > distance1:
                directionLeft = False
            else:
                directionLeft = True

        expand_count = int(distance_final / self.sidewiseline)
        self.leftExpand = expand_count if directionLeft is True else 0
        self.rightExpand = expand_count if directionLeft is not True else 0

    def calculate(self):
        startx = self.point_first[0]
        starty = self.point_first[1]
        endx = self.point_final[0]
        endy = self.point_final[1]
        if startx and starty and endx and endy:
            self._points = []
            self._lines = []
            lineStardEndPoints = []

            ########
            geod = pyproj.Geod(ellps="WGS84")
            angle, backAngle, distanceTmp = geod.inv(
                startx, starty, endx, endy)

            long = startx
            lat = starty
            for index in range(self.leftExpand):
                long, lat, tmpAngle = geod.fwd(
                    long, lat, angle-90, self.sidewiseline)
                e_long, e_lat, tempAngle = geod.fwd(
                    long, lat, angle, distanceTmp)
                lineStardEndPoints.append((long, lat, e_long, e_lat))

            lineStardEndPoints.append((startx, starty, endx, endy))

            long = startx
            lat = starty
            for index in range(self.rightExpand):
                long, lat, tmpAngle = geod.fwd(
                    long, lat, angle+90, self.sidewiseline)
                e_long, e_lat, tempAngle = geod.fwd(
                    long, lat, angle, distanceTmp)
                lineStardEndPoints.append((long, lat, e_long, e_lat))
            #######

            self.totalProgressValue = len(lineStardEndPoints)
            for line in lineStardEndPoints:
                linePointsResult, forwardAngle = self.caculate_line(*line)
                self._points.append(linePointsResult)
                self._lines.append(line)

                self.currentProgressValue = self.currentProgressValue + 1
                self.emit_progress("process the line:{}".format(
                    self.currentProgressValue))
            return True
        else:
            return False



"""
A simple photogrammetry routeing planner for UAV ,

Requirements: Python 3.6+.

Contact:  Xiangyong Luo <solo_lxy@126.com>

BSD 2-Clause License

Copyright (c) 2021, luoxiangyong
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
#import laspy
import numpy as np
import random
import pyproj
#from plyfile import PlyData, PlyElement

from ..core.alg  import *

class PlyExportor(SimpleExportor):
    def __init__(self,filename, aboveGround = 0.):
        super().__init__()
        self.name = "PLY Exporter"

        self.vertex = None

        self.x = []
        self.y = []
        self.z = []

        self.classification = []

        self.filename = filename

        self.outfile = None

        self.aboveGround = aboveGround
    
    def generatePointsLayer(self,calculator):
    
        # 写入点
        id = 0
        lineIndex = 0
        
        print("Height:", calculator.flight_height())
        for line in calculator.points:
            lineIndex = lineIndex + 1
            id = 0
            for p in line:
                id = id + 1
                name = "{}".format(id)
                lineName = "{}".format(lineIndex)
                self.x.append(p[0])
                self.y.append(p[1])
                self.z.append(self.aboveGround+calculator.flight_height())
                self.classification.append(5)

            calculator.set_progress_value(lineIndex, len(calculator.points), 
                            "Save ply file for points:{}".format(lineIndex))

    def generatePolygonsLayer(self, calculator):
        # 写入点对应的多边形
        idPolygon = 0
        lineIndex = 0
        for line in calculator.points:
            lineIndex = lineIndex + 1
            id = 0
            for p in line:
                id = id + 1
                name = "{}".format(id)
                lineName = "{}".format(lineIndex)
                rect  = self.calculateNoiseRectangle(p,calculator)

                self.x.append(rect[0][0])
                self.y.append(rect[0][1])
                self.z.append(self.aboveGround)

                self.classification.append(3)

                self.x.append(rect[1][0])
                self.y.append(rect[1][1])
                self.z.append(self.aboveGround)

                self.classification.append(3)

                self.x.append(rect[2][0])
                self.y.append(rect[2][1])
                self.z.append(self.aboveGround)

                self.classification.append(3)

                self.x.append(rect[3][0])
                self.y.append(rect[3][1])
                self.z.append(self.aboveGround)

                self.classification.append(2)
                

            calculator.set_progress_value(lineIndex, len(calculator.points), 
                            "Save las file for polygons:{}".format(lineIndex))

    def calculateNoiseRectangle(self,point,calculator):
        width = calculator.cameraWidth * calculator.gsd
        height = calculator.cameraHeight * calculator.gsd

        imgAngle = math.atan(calculator.cameraWidth*1.0/calculator.cameraHeight) * 180/math.pi
        randomAngle = random.randint(0,5)
        imgAngle = imgAngle + randomAngle

        geod = pyproj.Geod(ellps="WGS84")

        # 矩形的对角线长
        distance = math.sqrt(math.pow(width,2) + math.pow(height,2)) 

        #print("矩形的计算值：width={} height={} dj = {}".format(width,height,distance))

        # 计算右上角点
        angleTR = calculator.courseAngle - imgAngle
        longTR,latTR,tmpAngle = geod.fwd(point[0],point[1],angleTR, distance/2)

        # 计算右下角点
        angleBR = calculator.courseAngle + imgAngle 
        longBR,latBR,tmpAngle = geod.fwd(point[0],point[1],angleBR, distance/2)

        # 计算左下角点
        angleBL = angleTR + 180
        longBL,latBL,tmpAngle = geod.fwd(point[0],point[1],angleBL, distance/2)

        # 计算左上角点
        angleTL = angleBR + 180
        longTL,latTL,tmpAngle = geod.fwd(point[0],point[1],angleTL, distance/2)

        result = []
        result.append((longTR,latTR))
        result.append((longBR,latBR))
        result.append((longBL,latBL))
        result.append((longTL,latTL))
        # 多边形闭合
        result.append((longTR,latTR))

        return result

    def save(self,calculator):
        self.generatePointsLayer(calculator)
        self.generatePolygonsLayer(calculator)
        

        # 转换成投影坐标
        srcCRS = pyproj.Proj(proj = 'latlon', datum = 'WGS84')    # 定义数据地理坐标系
        targetCRS = pyproj.Proj(proj = 'merc', datum = 'WGS84')  # 定义转换投影坐标系
        for i in range(len(self.x)):
            
            #x1, y1 = srcCRS(self.x[i], self.y[i])
            x2, y2 = pyproj.transform(srcCRS, targetCRS, self.x[i], self.y[i])
            self.x[i] = x2
            self.y[i] = y2

            #print(x2,y2, self.z[i])

        points = [(self.x[i], self.y[i], self.z[i]) for i in range(len(self.x))]

        # face_data = np.array([([0, pIndex, pIndex + 1], 255, 255, 255),
        #                       ([0, pIndex, pIndex + 3], 255,   0,   0),
        #                      ],
        #             dtype=[('vertex_indices', 'i4', (3,)),
        #                    ('red', 'u1'), ('green', 'u1'),
        #                    ('blue', 'u1')])


        totalStations = 0
        for line in calculator.points:
            totalStations = totalStations + len(line)
        vetexCount = len(self.x)
        edgeCount = totalStations * 4
        faceCount = totalStations * 4
        rectCount = totalStations

        print("total station:", totalStations)
        print("The vetex count:", vetexCount)
        print("The edge count:", edgeCount)
        print("The tri face count:", faceCount)
        print("The rect face count:", rectCount)
        
        vertex = np.array(points,dtype=[('x', 'f8'), ('y', 'f8'),('z', 'f8')])
        print("vertex array shape:" , vertex.shape)

        with open(self.filename,'w') as file:
            file.write("""ply
format ascii 1.0
element vertex {}
comment vertices
property double x
property double y
property double z
property uint8 red
property uint8 green
property uint8 blue
element edge {}
property int32 vertex1
property int32 vertex2
property uint8 red
property uint8 green 
property uint8 blue 
element face {}
property list uchar int vertex_indices
end_header
""".format(vetexCount,edgeCount,faceCount+rectCount))

            # write vexex
            colors = np.zeros((totalStations,3),dtype=np.uint8)
            for i in range(len(self.x)):
                if i < totalStations:
                    colors[i,:] = random.randint(0,255),random.randint(0,255),random.randint(0,255)
                    file.write("{} {} {} {} {} {}\n".format(self.x[i], self.y[i], self.z[i],
                                *colors[i,:]))
                    file.flush()
                else:
                    index = int((i - totalStations)/4)
                    #print("对应的摄站点为:",index)
                    file.write("{} {} {} {} {} {}\n".format(self.x[i], self.y[i], self.z[i],
                        *colors[index,:]))
                    file.flush()

            # write edge
            for i in range(totalStations):
                pIndex = totalStations + i * 4
            
                file.write("{} {} 0 255 0\n".format(i, pIndex))
                file.flush()
                file.write("{} {} 0 255 0\n".format(i, pIndex + 1))
                file.flush()
                file.write("{} {} 0 255 0\n".format(i, pIndex + 2))
                file.flush()
                file.write("{} {} 0 255 0\n".format(i, pIndex + 3))
                file.flush()

            # write face
            for i in range(totalStations):
                
                pIndex = totalStations + i * 4
                #print("Current station:{}, pIndex:{}".format(i,pIndex))

                f1 = (i,  pIndex, pIndex + 1)
                f2 = (i , pIndex, pIndex + 3)
                f3 = (i , pIndex + 3, pIndex + 2)
                f4 = (i , pIndex + 2, pIndex + 1)

                #print("conic:", i, f1,f2,f3,f4)

                file.write("3 {} {} {}\n".format(f1[0], f1[1],f1[2]))
                file.flush()
                file.write("3 {} {} {}\n".format(f2[0], f2[1],f2[2]))
                file.flush()
                file.write("3 {} {} {}\n".format(f3[0], f3[1],f3[2]))
                file.flush()
                file.write("3 {} {} {}\n".format(f4[0], f4[1],f4[2]))
                file.flush()

            # write rect
            for i in range(totalStations):
                pIndex = totalStations + i * 4
            
                file.write("4 {} {} {} {}\n".format(pIndex, pIndex + 1,pIndex + 2, pIndex + 3))
                file.flush()

        return True

    # def save(self,calculator):
    #     self.generatePointsLayer(calculator)
    #     self.generatePolygonsLayer(calculator)
        

    #     # 转换成投影坐标
    #     srcCRS = pyproj.Proj(proj = 'latlon', datum = 'WGS84')    # 定义数据地理坐标系
    #     targetCRS = pyproj.Proj(proj = 'merc', datum = 'WGS84')  # 定义转换投影坐标系
    #     for i in range(len(self.x)):
            
    #         #x1, y1 = srcCRS(self.x[i], self.y[i])
    #         x2, y2 = pyproj.transform(srcCRS, targetCRS, self.x[i], self.y[i])
    #         self.x[i] = x2
    #         self.y[i] = y2

    #         #print(x2,y2, self.z[i])

    #     points = [(self.x[i], self.y[i], self.z[i]) for i in range(len(self.x))]
    #     vertex = np.array(points,
    #                 dtype=[('x', 'f8'), ('y', 'f8'),('z', 'f8')])

    #     print("line count:", len(calculator.points))
    #     print("point count perline:", len(calculator.points[0]))
    #     print("vertex size:" , vertex.shape)

    #     el = PlyElement.describe(vertex, 'vertex', comments=['vertices'])


    #     pIndex = len(calculator.points[0])
        
    #     # face_data = np.array([([0, pIndex, pIndex + 1], 255, 255, 255),
    #     #                       ([0, pIndex, pIndex + 3], 255,   0,   0),
    #     #                      ],
    #     #             dtype=[('vertex_indices', 'i4', (3,)),
    #     #                    ('red', 'u1'), ('green', 'u1'),
    #     #                    ('blue', 'u1')])

    #     lineCount = len(calculator.points)
    #     countPerLine = len(calculator.points[0])

    #     totalPoints = lineCount * countPerLine
    #     faceCount = totalPoints * 4

    #     face_data = np.zeros((faceCount,3),
    #                     dtype="i4"
    #                 )

    #     print("total point:", totalPoints)
    #     print("The face count:", faceCount)
    #     print("The face shape:", face_data.shape)

    #     for i in range(totalPoints):
            
    #         pIndex = countPerLine + i * 4
    #         print("Current line:1, pIndex:{}".format(i,pIndex))

    #         f1 = (i, pIndex, pIndex + 1)
    #         f2 = (i , pIndex, pIndex + 3)
    #         f3 = (i , pIndex + 3, pIndex + 2)
    #         f4 = (i , pIndex + 2, pIndex + 1)

    #         print("conic:", i, f1,f2,f3,f4)

    #         face_data[i] = f1
    #         face_data[i+1] = f2
    #         face_data[i+2] = f3
    #         face_data[i+3] = f4


    #     ply_faces = np.zeros(faceCount,
    #                      dtype=[('vertex_indices', 'i4', (3,))])

    #     ply_faces['vertex_indices'] = face_data

    #     face = PlyElement.describe(ply_faces,name='face')

    #     PlyData([el,face],text=True).write(self.filename)

    #     return True


from setuptools import setup, find_packages
import os

# Handle Version
def get_version(version_tuple):
    if not isinstance(version_tuple[-1], int):
        return '.'.join(
            map(str, version_tuple[:-1])
        ) + version_tuple[-1]

    return '.'.join(map(str, version_tuple))

init = os.path.join(os.path.dirname(__file__), 'sprp', '__init__.py')
version_line  = list(
        filter(lambda l: l.startswith("VERSION"), open(init))
    )[0]
VERSION = get_version(eval(version_line.split('=')[-1]))

def read_file(filename):
    basepath = os.path.dirname(os.path.dirname(__file__))
    filepath = os.path.join(basepath, filename)
    if os.path.exists(filepath):
        return open(filepath).read()
    else:
        return ''

setup(
    name    = 'sprp',
    version = VERSION,
    author  =  'Xiangyong Luo',
    long_description=read_file('README.txt'),
    author_email = 'solo_lxy@126.com',
    packages = find_packages(),
    include_package_data = True,
    url = 'https://github.com/luoxiangyong/sprp.git',
    keywords = ['UAV', 'photogrammetry','plan'],
    install_requires=[
        'shapely',
        'pyproj',
        'numpy',
        'gdal',
        'geojson',
        #'cython',
        'flask',
        'flask_cors'
    ],
    entry_points={
        "console_scripts":[
            'sprp-web=sprp.app.apiserver:main',
            'sprp-cmd=sprp.app.sprpcmd:main'
        ]
    }

)



"""
A simple photogrammetry routeing planner for UAV ,

Requirements: Python 3.6+.

Contact:  Xiangyong Luo <solo_lxy@126.com>

BSD 2-Clause License

Copyright (c) 2021, luoxiangyong
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
import laspy
import numpy as np
import random
import pyproj

from ..core.alg import *

class LasExportor(SimpleExportor):
    def __init__(self,filename, aboveGround = 0.):
        super().__init__()
        self.name = "Las Exporter"

        self.x = []
        self.y = []
        self.z = []

        self.classification = []

        self.filename = filename

        self.outfile = None

        self.aboveGround = aboveGround
    
    def generatePointsLayer(self,calculator):
    
        # 写入点
        id = 0
        lineIndex = 0
        
        print("Height:", calculator.flight_height())
        for line in calculator.points:
            lineIndex = lineIndex + 1
            id = 0
            for p in line:
                id = id + 1
                name = "{}".format(id)
                lineName = "{}".format(lineIndex)
                self.x.append(p[0])
                self.y.append(p[1])
                self.z.append(self.aboveGround+calculator.flight_height())
                self.classification.append(5)

            calculator.set_progress_value(lineIndex, len(calculator.points), 
                            "Save las file for points:{}".format(lineIndex))

    def generatePolygonsLayer(self, calculator):
        # 写入点对应的多边形
        idPolygon = 0
        lineIndex = 0
        for line in calculator.points:
            lineIndex = lineIndex + 1
            id = 0
            for p in line:
                id = id + 1
                name = "{}".format(id)
                lineName = "{}".format(lineIndex)
                rect  = self.calculateNoiseRectangle(p,calculator)

                self.x.append(rect[0][0])
                self.y.append(rect[0][1])
                self.z.append(self.aboveGround)

                self.classification.append(3)

                self.x.append(rect[1][0])
                self.y.append(rect[1][1])
                self.z.append(self.aboveGround)

                self.classification.append(3)

                self.x.append(rect[2][0])
                self.y.append(rect[2][1])
                self.z.append(self.aboveGround)

                self.classification.append(3)

                self.x.append(rect[3][0])
                self.y.append(rect[3][1])
                self.z.append(self.aboveGround)

                self.classification.append(2)
                

            calculator.set_progress_value(lineIndex, len(calculator.points), 
                            "Save las file for polygons:{}".format(lineIndex))

    def calculateNoiseRectangle(self,point,calculator):
        width = calculator.cameraWidth * calculator.gsd
        height = calculator.cameraHeight * calculator.gsd

        imgAngle = math.atan(calculator.cameraWidth*1.0/calculator.cameraHeight) * 180/math.pi
        randomAngle = random.randint(0,5)
        imgAngle = imgAngle + randomAngle

        geod = pyproj.Geod(ellps="WGS84")

        # 矩形的对角线长
        distance = math.sqrt(math.pow(width,2) + math.pow(height,2)) 

        #print("矩形的计算值：width={} height={} dj = {}".format(width,height,distance))

        # 计算右上角点
        angleTR = calculator.courseAngle - imgAngle
        longTR,latTR,tmpAngle = geod.fwd(point[0],point[1],angleTR, distance/2)

        # 计算右下角点
        angleBR = calculator.courseAngle + imgAngle 
        longBR,latBR,tmpAngle = geod.fwd(point[0],point[1],angleBR, distance/2)

        # 计算左下角点
        angleBL = angleTR + 180
        longBL,latBL,tmpAngle = geod.fwd(point[0],point[1],angleBL, distance/2)

        # 计算左上角点
        angleTL = angleBR + 180
        longTL,latTL,tmpAngle = geod.fwd(point[0],point[1],angleTL, distance/2)

        result = []
        result.append((longTR,latTR))
        result.append((longBR,latBR))
        result.append((longBL,latBL))
        result.append((longTL,latTL))
        # 多边形闭合
        result.append((longTR,latTR))

        return result

    def save(self,calculator):
        hdr = laspy.header.Header()
        self.outfile = laspy.file.File(self.filename, mode="w", header=hdr)

        self.generatePolygonsLayer(calculator)
        self.generatePointsLayer(calculator)

        srcCRS = pyproj.Proj(proj = 'latlon', datum = 'WGS84')    # 定义数据地理坐标系
        targetCRS = pyproj.Proj(proj = 'merc', datum = 'WGS84')  # 定义转换投影坐标系
        for i in range(len(self.x)):
            
            #x1, y1 = srcCRS(self.x[i], self.y[i])
            x2, y2 = pyproj.transform(srcCRS, targetCRS, self.x[i], self.y[i])
            self.x[i] = x2
            self.y[i] = y2

            print(x2,y2, self.z[i])

        allx = np.array(self.x) # Four Points
        ally = np.array(self.y)
        allz = np.array(self.z)

        xmin = np.floor(np.min(allx))
        ymin = np.floor(np.min(ally))
        zmin = np.floor(np.min(allz))

        self.outfile.header.offset = [xmin,ymin,zmin]
        self.outfile.header.scale = [0.001,0.001,0.001]

        self.outfile.x = allx
        self.outfile.y = ally
        self.outfile.z = allz
        self.outfile.classification_flags = self.classification
        self.outfile.pt_src_id = np.linspace(0,len(allx))
        

        self.outfile.close()

        print("line count:", len(calculator.points))
        print("point count perline:", len(calculator.points[0]))
        print("total point count:", len(allx))

        print(allz)

        return True